@@ -25,24 +25,22 @@
 #include <gtk/gtk.h>
 #include <gdk/gdkx.h>
 
-#include <libindicate/server.h>
-#include <libindicate/indicator.h>
-#include <libindicate-gtk/indicator.h>
 #include <telepathy-glib/telepathy-glib.h>
 #include <telepathy-glib/simple-approver.h>
 
 #include <unity.h>
+#include <messaging-menu.h>
 
-static GHashTable *indicators = NULL;
-static gboolean has_interest = FALSE;
+static MessagingMenuApp *mmapp;
+GHashTable *dispatch_ops;
+static gint n_sources;
 UnityLauncherEntry *launcher = NULL;
 GList *conns;
 
-static void update_launcher ()
+static void update_launcher (gint count)
 {
 	if (launcher != NULL)
 	{
-		gint count = g_hash_table_size (indicators);
 		g_debug ("unity launcher: count is %d", count);
 		if (count > 0)
 		{
@@ -57,115 +55,90 @@ static void update_launcher ()
 	}
 }
 
-static IndicateIndicator *
-find_indicator_by_id (const gchar *id)
+static gchar *
+channel_id (TpChannel *channel)
 {
-	if (!check_indicator (id))
-	{
-		return NULL;
-	}
+	TpConnection *con;
 
-	const gchar *channel_id;
-	g_debug ("find_indicator_by_id: %s", id);
+	con = tp_channel_borrow_connection (channel);
 
-	GHashTableIter iter;
-	gpointer data, indicator;
-
-	g_hash_table_iter_init (&iter, indicators);
-	while (g_hash_table_iter_next (&iter, &indicator, &data)) {
-		channel_id = indicate_indicator_get_property (INDICATE_INDICATOR (indicator), "channel-id");
-		g_debug ("indicator for channel %s", channel_id);
-		if (g_strcmp0 (id, channel_id) == 0)
-		{
-			return INDICATE_INDICATOR (indicator);
-		}
-	}
-
-	return NULL;
+	return g_strconcat ("channel:",
+			    tp_proxy_get_object_path (TP_PROXY (con)), ":",
+			    tp_proxy_get_object_path (TP_PROXY (channel)),
+			    NULL);
 }
 
-static void 
-indicator_remove_by_indicator (IndicateIndicator *indicator)
+static gchar *
+contact_id (TpContact *contact)
 {
-	g_debug ("indicator_remove_by_indicator");
-	if (indicator != NULL)
-	{
-		if (g_hash_table_lookup (indicators, indicator) != NULL)
-		{
-			g_debug ("removing indicator");
-			indicate_indicator_hide (INDICATE_INDICATOR (indicator));
-			g_hash_table_remove (indicators, indicator);
-		}
-		update_launcher ();
-		g_object_unref (indicator);
-	}
+	TpConnection *con;
+
+	con = tp_contact_get_connection (contact);
+
+	return g_strconcat ("contact:",
+			    tp_proxy_get_object_path (TP_PROXY (con)), ":",
+			    tp_contact_get_identifier (contact),
+			    NULL);
 }
 
 static void
-indicator_remove_by_id (const gchar *id)
+dispatch_op_finished_cb (TpProxy *self, guint domain, gint code, gchar *message, gpointer user_data)
 {
-	g_debug ("indicator_remove_by_id: %s", id);
-	IndicateIndicator *indicator = find_indicator_by_id (id);
+	TpChannelDispatchOperation *dispatch_op = TP_CHANNEL_DISPATCH_OPERATION (self);
+	GPtrArray *channels;
 
-	if (indicator != NULL)
+	channels = tp_channel_dispatch_operation_borrow_channels (dispatch_op);
+	if (channels != NULL && channels->len)
 	{
-		indicator_remove_by_indicator (INDICATE_INDICATOR (indicator));
+		TpChannel *channel;
+		gchar *id;
+
+		channel = g_ptr_array_index (channels, 0);
+		id = channel_id (channel);
+
+		messaging_menu_app_remove_source (mmapp, id);
+		g_hash_table_remove (dispatch_ops, id);
+		update_launcher (n_sources--);
+
+		g_free (id);
 	}
 }
 
-
-static void
-dispatch_op_finished_cb (TpProxy *self, guint domain, gint code, gchar *message, IndicateIndicator *indicator)
-{
-	g_return_if_fail (INDICATE_IS_INDICATOR(indicator));
-
-	g_debug ("DispatchOperation invalidated, removing indicator");
-	indicator_remove_by_indicator (INDICATE_INDICATOR (indicator));
-}
-
 static void
 handle_with_cb (GObject *self, GAsyncResult *result, gpointer user_data)
 {
-	IndicateIndicator *indicator = INDICATE_INDICATOR (user_data);
-	g_return_if_fail (INDICATE_IS_INDICATOR(indicator));
-
 	TpChannelDispatchOperation *dispatch_op = TP_CHANNEL_DISPATCH_OPERATION (self);
 	GError *error = NULL;
+	TpChannel *channel;
+	GPtrArray *channels;
+	gchar *id;
 
+	channels = tp_channel_dispatch_operation_borrow_channels (dispatch_op);
+	channel = g_ptr_array_index (channels, 0);
+	id = channel_id (channel);
 
-	if (!tp_channel_dispatch_operation_handle_with_time_finish (dispatch_op, result, &error))
+	if (!tp_channel_dispatch_operation_handle_with_finish (dispatch_op, result, &error))
 	{
                 if (error)
 		{
 			g_warning ("Failed to handle operations: %s\n", error->message);
 			g_error_free (error);
-			g_debug ("Activated, removing indicator");
-			indicator_remove_by_indicator (INDICATE_INDICATOR (indicator));
+			messaging_menu_app_remove_source (mmapp, id);
+			update_launcher (n_sources--);
 		}
 	}
-}
 
-
-static void
-indicator_display (IndicateIndicator *indicator, guint timestamp, TpChannelDispatchOperation * dispatch_op)
-{
-	g_debug ("indicator_display");
-	tp_channel_dispatch_operation_handle_with_time_async (dispatch_op, NULL, timestamp, handle_with_cb, indicator);
+	g_free (id);
 }
 
 static void
 observer_handle_with_cb (GObject *self, GAsyncResult *result, gpointer user_data)
 {
 	g_debug ("observer_handle_with_cb");
-	IndicateIndicator *indicator = INDICATE_INDICATOR (user_data);
-	g_return_if_fail (INDICATE_IS_INDICATOR(indicator));
 
 	TpChannelDispatcher *cd = TP_CHANNEL_DISPATCHER (self);
 	GError *error = NULL;
 
-	g_debug ("Activated, removing indicator");
-	indicator_remove_by_indicator (INDICATE_INDICATOR (indicator));
-	
 	tp_channel_dispatcher_present_channel_finish (cd, result, &error);
 	if (error)
 	{
@@ -174,411 +147,6 @@ observer_handle_with_cb (GObject *self, GAsyncResult *result, gpointer user_data
 	}
 }
 
-static void
-observer_indicator_display (IndicateIndicator *indicator, guint timestamp, gpointer user_data)
-{
-        g_debug ("observer_indicator_display");
-	TpChannel *channel = TP_CHANNEL (user_data);
-	g_return_if_fail (TP_IS_CHANNEL (channel));
-
-	g_object_ref (indicator);
-	GError *error = NULL;
-
-	TpDBusDaemon *bus = tp_dbus_daemon_dup (&error);
-
-	TpChannelDispatcher *cd = tp_channel_dispatcher_new (bus);
-	g_object_unref (bus);
-
-	if (error)
-	{
-		g_warning ("Failed to get Daemon: %s\n", error->message);
-		g_error_free (error);
-	}
-
-	tp_channel_dispatcher_present_channel_async (cd, TP_CHANNEL (channel), timestamp, observer_handle_with_cb, INDICATE_INDICATOR (indicator));
-}
-
-gboolean check_indicator (const gchar *id)
-{
-	const gchar *channel_id;
-	g_debug ("check_indicator: %s", id);
-	GHashTableIter iter;
-	gpointer data, ind;
-
-	g_hash_table_iter_init (&iter, indicators);
-	while (g_hash_table_iter_next (&iter, &ind, &data)) {
-		channel_id = indicate_indicator_get_property (INDICATE_INDICATOR (ind), "channel-id");
-		g_debug ("indicator for channel %s", channel_id);
-		if (g_strcmp0 (id, channel_id) == 0)
-		{
-			return TRUE;
-		}
-	}
-
-	return FALSE;
-}
-
-static TpContact *
-get_contact_from_channel (TpChannel * channel)
-{
-	g_return_val_if_fail (TP_IS_CHANNEL (channel), NULL);
-	TpContact *contact = NULL;
-	if (tp_channel_get_requested (TP_CHANNEL (channel)))
-		contact = tp_channel_get_target_contact (TP_CHANNEL (channel));
-	else
-		contact = tp_channel_get_initiator_contact (TP_CHANNEL (channel));
-	return TP_CONTACT (contact);
-}
-	
-
-static void
-handle_contacts_pending_add_indicator_cb (TpConnection *connection, guint n_contacts, TpContact * const *contacts, guint n_failed, const TpHandle *failed, const GError *err, gpointer user_data, GObject *weak_object)
-{
-	g_debug ("handle_contacts_pending_add_indicator_cb");
-	TpChannel *channel = TP_CHANNEL (user_data);
-	GError *error = NULL;
-	const gchar *id = "";
-	const gchar *alias = "";
-	GFile *avatar = NULL;
-	GdkPixbuf *pixbuf = NULL;
-	TpContact *contact = NULL;
-
-	contact = get_contact_from_channel (channel);
-
-	if (contact)
-	{
-		alias = tp_contact_get_alias (contact);
-		avatar = tp_contact_get_avatar_file (contact);
-		g_debug ("handle_contacts_pending_add_indicator_cb HAS CONTACT alias: %s", alias);
-	} else
-	{
-		alias = tp_channel_get_initiator_identifier (channel);
-		g_debug ("handle_contacts_pending_add_indicator_cb NO CONTACT alias: %s", alias);
-	}
-	g_debug ("handle_contacts_pending_add_indicator_cb alias: %s", alias);
-
-	id = tp_channel_get_identifier (TP_CHANNEL (channel));
-	IndicateIndicator *indicator;
-	indicator = indicate_indicator_new ();
-	indicate_indicator_set_property (INDICATE_INDICATOR (indicator),
-					"subtype", "im");
-	indicate_indicator_set_property (INDICATE_INDICATOR (indicator),
-					"sender", alias);
-	indicate_indicator_set_property (INDICATE_INDICATOR (indicator),
-					"draw-attention", "true");
-	indicate_indicator_set_property (INDICATE_INDICATOR (indicator),
-					"channel-id", id);
-	g_debug ("added channel-id: %s", id);
-	g_signal_connect (G_OBJECT (indicator),
-		INDICATE_INDICATOR_SIGNAL_DISPLAY,
-		G_CALLBACK (observer_indicator_display), channel);
-
-	GTimeVal time;
-	g_get_current_time (&time);
-	indicate_indicator_set_property_time (INDICATE_INDICATOR (indicator),
-						"time", &time);
-	if (avatar)
-	{
-		pixbuf = gdk_pixbuf_new_from_file_at_scale (g_file_get_path (avatar), 22, 22, TRUE, &error);
-	}
-
-	if (error)
-	{
-		g_warning ("Failed to create pixbuf: %s", error->message);
-		g_error_free (error);
-	}
-
-	if (pixbuf)
-	{
-		indicate_gtk_indicator_set_property_icon (indicator, "icon", pixbuf);
-		g_object_unref(G_OBJECT(pixbuf));
-	}
-
-	indicate_indicator_show (INDICATE_INDICATOR (indicator));
-	g_hash_table_insert(indicators, indicator, channel);
-	update_launcher ();
-}
-
-gboolean check_pending_messages (gpointer data)
-{
-	TpTextChannel *channel = TP_TEXT_CHANNEL (data);
-	TpConnection *connection;
-	GList *pending;
-	const gchar *id;
-	gboolean has_indicator = FALSE;
-	TpHandle handle;
-
-	static TpContactFeature features[] = {
-		TP_CONTACT_FEATURE_ALIAS,
-		TP_CONTACT_FEATURE_AVATAR_DATA,
-	};
-
-	pending = tp_text_channel_get_pending_messages (TP_TEXT_CHANNEL (channel));
-	g_debug ("pending: %u", g_list_length (pending));
-	if (g_list_length (pending) > 0)
-	{
-		id = tp_channel_get_identifier (TP_CHANNEL (channel));
-
-		if (!check_indicator (id))
-		{
-			connection = tp_channel_borrow_connection (TP_CHANNEL (channel));
-			handle = tp_channel_get_initiator_handle (TP_CHANNEL (channel));
-			tp_connection_get_contacts_by_handle (connection, 1, &handle, 2, features, handle_contacts_pending_add_indicator_cb, channel, NULL, NULL);
-		}
-	}
-        g_list_free (pending);
-	return FALSE;
-}
-
-static void
-pending_message_removed_cb (TpTextChannel *channel, TpMessage *message, gpointer user_data)
-{
-	const gchar *id;
-	g_debug ("pending_message_removed_cb: %s", tp_channel_get_identifier (TP_CHANNEL (channel)));
-	id = tp_channel_get_identifier (TP_CHANNEL (channel));
-	indicator_remove_by_id (id);
-}
-
-static void
-message_received_cb (TpTextChannel *channel, TpMessage *message, gpointer user_data)
-{
-	g_debug ("message_received_cb: %s", tp_channel_get_identifier (TP_CHANNEL (channel)));
-	g_timeout_add (100, check_pending_messages, channel);
-}
-
-static void
-handle_contacts_add_indicator_cb (TpConnection *connection, guint n_contacts, TpContact * const *contacts, guint n_failed, const TpHandle *failed, const GError *err, gpointer user_data, GObject *weak_object)
-{
-	IndicateIndicator *indicator = NULL;
-	TpContact *contact = NULL;
-	const gchar *id = "";
-	const gchar *alias = "";
-	GTimeVal time;
-	GFile *avatar = NULL;
-	GdkPixbuf *pixbuf = NULL;
-	GError *error = NULL;
-	
-	TpChannelDispatchOperation *dispatch_op = TP_CHANNEL_DISPATCH_OPERATION (user_data);
-
-	TpChannel *channel;
-	GPtrArray *channels;
-	channels = tp_channel_dispatch_operation_borrow_channels (dispatch_op);
-	channel = g_ptr_array_index (channels, 0);
-
-	contact = get_contact_from_channel (channel);
-
-	if (contact)
-	{
-		alias = tp_contact_get_alias (contact);
-		avatar = tp_contact_get_avatar_file (contact);
-	} else
-	{
-		alias = tp_channel_get_initiator_identifier (channel);
-	}
-	g_debug ("handle_contacts_add_indicator_cb alias: %s", alias);
-
-	id = tp_channel_get_identifier (TP_CHANNEL (channel));
-
-	if (check_indicator (id))
-	{
-		indicator_remove_by_id (id);
-	}
-
-	indicator = indicate_indicator_new ();
-	indicate_indicator_set_property (INDICATE_INDICATOR (indicator),
-					 "subtype", "im");
-	indicate_indicator_set_property (INDICATE_INDICATOR (indicator),
-					 "sender", alias);
-	indicate_indicator_set_property (INDICATE_INDICATOR (indicator),
-					 "draw-attention", "true");
-	indicate_indicator_set_property (INDICATE_INDICATOR (indicator),
-					 "channel-id", id);
-	g_debug ("added channel-id: %s", id);
-
-	if (avatar)
-	{
-		pixbuf = gdk_pixbuf_new_from_file_at_scale (g_file_get_path (avatar), 22, 22, TRUE, &error);
-	}
-
-	if (error)
-	{
-		g_warning ("Failed to create pixbuf: %s", error->message);
-		g_error_free (error);
-	}
-
-	if (pixbuf)
-	{
-		indicate_gtk_indicator_set_property_icon (indicator, "icon", pixbuf);
-		g_object_unref(G_OBJECT(pixbuf));
-	}
-
-	g_signal_connect (G_OBJECT (indicator),
-			INDICATE_INDICATOR_SIGNAL_DISPLAY,
-			G_CALLBACK (indicator_display), dispatch_op);
-
-	g_get_current_time (&time);
-	indicate_indicator_set_property_time (INDICATE_INDICATOR (indicator),
-					"time", &time);
-	indicate_indicator_show (INDICATE_INDICATOR (indicator));
-	g_hash_table_insert(indicators, indicator, dispatch_op);
-	update_launcher ();
-	g_signal_connect (dispatch_op, "invalidated", G_CALLBACK (dispatch_op_finished_cb), indicator);
-}
-
-static void
-add_dispatch_operation_cb (TpSimpleApprover *self, TpAccount *account, TpConnection *connection,
-    GList *channels, TpChannelDispatchOperation *dispatch_op, TpAddDispatchOperationContext *context,
-    gpointer user_data)
-{
-	if (has_interest == FALSE)
-	{
-		g_debug ("add_dispatch_operation_cb but there are no listeners with interest");
-		tp_add_dispatch_operation_context_accept (context);
-		return;
-	}
-
-	GList *itr;
-	TpChannel *channel;
-	TpHandle handle;
-
-	static TpContactFeature features[] = {
-		TP_CONTACT_FEATURE_ALIAS,
-		TP_CONTACT_FEATURE_AVATAR_DATA,
-	};
-
-	g_object_ref (dispatch_op);
-	for (itr = channels; itr != NULL; itr = g_list_next (itr))
-	{
-		channel = itr->data;
-		handle = tp_channel_get_initiator_handle (channel);
-		tp_connection_get_contacts_by_handle (connection, 1, &handle, 2, features, handle_contacts_add_indicator_cb, dispatch_op, NULL, NULL);
-	}
-	g_list_free(itr);
-	tp_add_dispatch_operation_context_accept (context);
-}
-
-static void
-observer_add_dispatch_operation_cb (TpSimpleObserver *self, TpAccount *account, TpConnection *connection,
-    GList *channels, TpChannelDispatchOperation *dispatch_op, GList *requests, TpObserveChannelsContext *context,
-    gpointer user_data)
-{
-	if (has_interest == FALSE)
-	{
-		g_debug ("observer_add_dispatch_operation_cb but there are no listeners with interest");
-		tp_observe_channels_context_accept (context);
-		return;
-	}
-	GList *itr;
-	TpChannel *channel;
-
-	for (itr = channels; itr != NULL; itr = g_list_next (itr))
-	{
-		channel = itr->data;
-		if (TP_IS_TEXT_CHANNEL (channel))
-		{
-			g_debug ("Handling text channel: %s", tp_channel_get_identifier (channel));
-
-			g_object_ref (channel);
-
-			g_signal_connect (channel, "message-received",
-				G_CALLBACK (message_received_cb), NULL);
-			g_signal_connect (channel, "pending-message-removed",
-				G_CALLBACK (pending_message_removed_cb), NULL);
-
-			check_pending_messages (TP_CHANNEL (channel));
-		}
-	}
-	g_list_free(itr);
-	tp_observe_channels_context_accept (context);
-}
-
-static void
-server_display (IndicateServer * indicate_server, guint timestamp, gpointer user_data)
-{
-	g_debug ("server_display at %u", timestamp);
-
-	if (has_interest == FALSE)
-	{
-		g_debug ("server_display but there are no listeners with interest");
-		return;
-	}
-
-	GDesktopAppInfo *app;
-	GError *error = NULL;
-
-	g_unsetenv ("DESKTOP_AUTOSTART_ID");
-
-	app = g_desktop_app_info_new ("empathy.desktop");
-	g_app_info_launch ((GAppInfo*)app, NULL, NULL, &error);
-
-	if (error)
-	{
-		g_warning ("Failed to launch empathy: %s", error->message);
-		g_error_free (error);
-	}
-
-	g_object_unref (app);
-}
-
-static void
-indicate_server_show_interest (IndicateServer * server, IndicateInterests interest, gpointer data)
-{
-	g_debug ("indicate_server_show_interest");
-	if (interest == INDICATE_INTEREST_SERVER_SIGNAL)
-	{
-		if (has_interest == FALSE)
-		{
-			has_interest = TRUE;
-		}
-	}
-	return;
-}
-
-static void
-indicate_server_remove_interest (IndicateServer * server, IndicateInterests interest, gpointer data)
-{
-	g_debug ("indicate_server_remove_interest");
-	if (interest == INDICATE_INTEREST_SERVER_SIGNAL)
-	{
-		if (has_interest == TRUE)
-		{
-			has_interest = FALSE;
-		}
-	}
-	return;
-}
-
-static IndicateServer *
-indicate_server_setup ()
-{
-	IndicateServer *indicate_server = NULL;
-
-	const gchar *desktop_file = "/usr/share/applications/empathy.desktop";
-	indicate_server = indicate_server_ref_default ();
-	indicate_server_set_type (indicate_server, "message.im");
-
-	indicate_server_set_desktop_file (indicate_server, desktop_file);
-	indicate_server_show (indicate_server);
-	indicators = g_hash_table_new (g_direct_hash, g_direct_equal);
-	launcher = unity_launcher_entry_get_for_desktop_id ("empathy.desktop");
-
-	g_signal_connect (G_OBJECT (indicate_server),
-		INDICATE_SERVER_SIGNAL_SERVER_DISPLAY,
-		G_CALLBACK (server_display), NULL);
-
-	g_signal_connect (G_OBJECT (indicate_server), 
-		INDICATE_SERVER_SIGNAL_INTEREST_ADDED, 
-		G_CALLBACK (indicate_server_show_interest), NULL);
-	g_signal_connect (G_OBJECT (indicate_server), 
-		INDICATE_SERVER_SIGNAL_INTEREST_REMOVED, 
-		G_CALLBACK (indicate_server_remove_interest), NULL);
-
-	has_interest = indicate_server_check_interest (indicate_server, 
-				INDICATE_INTEREST_SERVER_SIGNAL);
-	
-	return indicate_server;
-}
-
 static void
 contact_request_subscription_cb (GObject *source, GAsyncResult *result, gpointer user_data)
 { 
@@ -669,10 +237,14 @@ contact_add_response_cb (GtkWidget *dialog, gint response, TpContact *contact)
 }
 
 static void
-handle_contacts_subscription_add_indicator_cb (TpConnection *connection, guint n_contacts, TpContact * const *contacts, const GError *err, gpointer user_data, GObject *weak_object)
+handle_contacts_subscription_add_indicator_cb (GObject     *object,
+					       GAsyncResult *result,
+					       gpointer      user_data)
 {
+	TpContact *contact;
 	GtkWidget *dialog;
 	GtkWidget *button;
+	gchar *id;
 	const gchar *alias = "";
 	gchar *title;
 	const gchar *message;
@@ -681,15 +253,21 @@ handle_contacts_subscription_add_indicator_cb (TpConnection *connection, guint n
 	GtkWidget *image = NULL;
 	GError *error = NULL;
 
-	IndicateIndicator *indicator = INDICATE_INDICATOR (user_data);
+	contact = tp_connection_dup_contact_by_id_finish (TP_CONNECTION (object), result, &error);
+	if (contact == NULL)
+	{
+		g_warning ("unable to find contact: %s", error->message);
+		g_error_free (error);
+		return;
+	}
 
-	message = tp_contact_get_publish_request (contacts[0]);
+	message = tp_contact_get_publish_request (contact);
 
-	alias = tp_contact_get_alias (contacts[0]);
+	alias = tp_contact_get_alias (contact);
 
 	g_debug ("Getting avatar for %s", alias);
 
-	avatar = tp_contact_get_avatar_file (contacts[0]);
+	avatar = tp_contact_get_avatar_file (contact);
 
 	if (avatar)
 	{
@@ -729,39 +307,357 @@ handle_contacts_subscription_add_indicator_cb (TpConnection *connection, guint n
 
 	button = gtk_dialog_add_button (GTK_DIALOG (dialog), _("_Accept"), GTK_RESPONSE_ACCEPT);
 
-	g_signal_connect (dialog, "response", G_CALLBACK (contact_add_response_cb), contacts[0]);
+	g_signal_connect (dialog, "response", G_CALLBACK (contact_add_response_cb), contact);
 
 	gtk_widget_show (GTK_WIDGET (dialog));
 
-	if (indicator != NULL)
+	id = contact_id (contact);
+	messaging_menu_app_remove_source (mmapp, id);
+
+	g_free (id);
+}
+
+static TpConnection *
+get_tp_connection (TpAccountManager *tp_am,
+		    const gchar      *object_path)
+{
+	TpSimpleClientFactory *factory;
+	TpConnection *con;
+	GError *error = NULL;
+
+	factory = tp_proxy_get_factory (tp_am);
+
+	con = tp_simple_client_factory_ensure_connection (factory, object_path, NULL, &error);
+	if (con == NULL)
 	{
-		if (g_hash_table_lookup (indicators, indicator) != NULL)
-		{
-			g_debug ("Activated, removing indicator");
-			indicate_indicator_hide (INDICATE_INDICATOR (indicator));
-			g_hash_table_remove (indicators, indicator);
-		}
-		g_object_unref (indicator);
+		g_warning ("cannot find telepathy connection '%s': %s",
+			   object_path, error->message);
+		g_error_free (error);
 	}
+
+	return con;
+}
+
+static TpChannel *
+get_tp_channel (TpAccountManager *tp_am,
+		 const gchar *connection_object_path,
+		 const gchar *channel_object_path)
+{
+	TpSimpleClientFactory *factory;
+	TpConnection *con;
+	TpChannel *channel;
+	GError *error = NULL;
+
+	factory = tp_proxy_get_factory (tp_am);
+
+	con = get_tp_connection (tp_am, connection_object_path);
+	if (con == NULL)
+		return NULL;
+
+	channel = tp_simple_client_factory_ensure_channel (factory, con, channel_object_path, NULL, &error);
+	if (channel == NULL)
+	{
+		g_warning ("cannot find telepathy channel '%s': %s",
+			   channel_object_path, error->message);
+		g_error_free (error);
+	}
+
+	g_object_unref (con);
+	return channel;
+}
+
+static TpChannelDispatcher *
+get_tp_channel_dispatcher ()
+{
+	TpDBusDaemon *bus;
+	TpChannelDispatcher *dispatcher;
+	GError *error = NULL;
+
+	bus = tp_dbus_daemon_dup (&error);
+	if (bus == NULL)
+	{
+		g_warning ("Failed to get Daemon: %s\n", error->message);
+		g_error_free (error);
+		return NULL;
+	}
+
+	dispatcher = tp_channel_dispatcher_new (bus);
+
+	g_object_unref (bus);
+	return dispatcher;
 }
 
 static void
-subscription_indicator_display (IndicateIndicator *indicator, guint timestamp, TpContact * contact)
+message_source_activated (MessagingMenuApp *app,
+			  const gchar      *id,
+			  gpointer          user_data)
 {
-	if (TP_IS_CONTACT (contact))
+	TpAccountManager *tp_am = user_data;
+	gchar **sections;
+
+	/* message source ids contain the type of source, the
+	 * tpconnection's object path, and the object path of the actual
+	 * source object, separated by colons */
+	sections = g_strsplit (id, ":", 0);
+	if (g_strv_length (sections) != 3)
+		goto out;
+
+	if (g_str_equal (sections[0], "channel"))
 	{
-		TpConnection *connection = tp_contact_get_connection (contact);
+		TpChannelDispatchOperation *dispatch_op;
+
+		/* accept dispatch operation if there is one for this channel */
+		dispatch_op = g_hash_table_lookup (dispatch_ops, id);
+		if (dispatch_op != NULL)
+		{
+			tp_channel_dispatch_operation_handle_with_async (dispatch_op, NULL, handle_with_cb, NULL);
+			g_hash_table_remove (dispatch_ops, id);
+		}
+		else
+		{
+			TpChannel *channel;
+			TpChannelDispatcher *dispatcher;
+
+			channel = get_tp_channel (tp_am, sections[1], sections[2]);
+			dispatcher = get_tp_channel_dispatcher ();
+			if (channel != NULL && dispatcher != NULL)
+			{
+				tp_channel_dispatcher_present_channel_async (dispatcher, channel,
+									     g_get_real_time () / G_USEC_PER_SEC,
+									     observer_handle_with_cb, NULL);
+			}
+
+			g_clear_object (&channel);
+			g_clear_object (&dispatcher);
+		}
+	}
+	else if (g_str_equal (sections[0], "contact"))
+	{
+		TpConnection *con;
+
 		static TpContactFeature features[] = {
 			TP_CONTACT_FEATURE_ALIAS,
 			TP_CONTACT_FEATURE_AVATAR_DATA,
 		};
 
-		TpContact *contacts[] = {
-			contact,
-		};
-
-		tp_connection_upgrade_contacts (connection, 1, contacts, 2, features, handle_contacts_subscription_add_indicator_cb, indicator, NULL, NULL);
+		con = get_tp_connection (tp_am, sections[1]);
+		if (con != NULL)
+		{
+			tp_connection_dup_contact_by_id_async (con, sections[2], 2, features,
+							       handle_contacts_subscription_add_indicator_cb, NULL);
+			g_object_unref (con);
+		}
 	}
+
+out:
+	g_strfreev (sections);
+}
+
+static TpContact *
+get_contact_from_channel (TpChannel * channel)
+{
+	g_return_val_if_fail (TP_IS_CHANNEL (channel), NULL);
+	TpContact *contact = NULL;
+	if (tp_channel_get_requested (TP_CHANNEL (channel)))
+		contact = tp_channel_get_target_contact (TP_CHANNEL (channel));
+	else
+		contact = tp_channel_get_initiator_contact (TP_CHANNEL (channel));
+	return TP_CONTACT (contact);
+}
+	
+
+static void
+handle_contacts_pending_add_indicator_cb (TpConnection *connection, guint n_contacts, TpContact * const *contacts, guint n_failed, const TpHandle *failed, const GError *err, gpointer user_data, GObject *weak_object)
+{
+	g_debug ("handle_contacts_pending_add_indicator_cb");
+	TpChannel *channel = TP_CHANNEL (user_data);
+	gchar *id;
+	const gchar *alias = "";
+	GFile *avatar_file;
+	GIcon *avatar = NULL;
+	TpContact *contact = NULL;
+
+	contact = get_contact_from_channel (channel);
+
+	if (contact)
+	{
+		alias = tp_contact_get_alias (contact);
+		avatar_file = tp_contact_get_avatar_file (contact);
+		avatar = avatar_file ? g_file_icon_new (avatar_file) : NULL;
+		g_debug ("handle_contacts_pending_add_indicator_cb HAS CONTACT alias: %s", alias);
+	} else
+	{
+		alias = tp_channel_get_initiator_identifier (channel);
+		g_debug ("handle_contacts_pending_add_indicator_cb NO CONTACT alias: %s", alias);
+	}
+	g_debug ("handle_contacts_pending_add_indicator_cb alias: %s", alias);
+
+	id = channel_id (channel);
+	messaging_menu_app_append_source (mmapp, id, avatar, alias);
+	messaging_menu_app_draw_attention (mmapp, id);
+
+	update_launcher (++n_sources);
+
+	g_clear_object (&avatar);
+	g_free (id);
+}
+
+gboolean check_pending_messages (gpointer data)
+{
+	TpTextChannel *channel = TP_TEXT_CHANNEL (data);
+	TpConnection *connection;
+	GList *pending;
+	gboolean has_indicator = FALSE;
+	TpHandle handle;
+
+	static TpContactFeature features[] = {
+		TP_CONTACT_FEATURE_ALIAS,
+		TP_CONTACT_FEATURE_AVATAR_DATA,
+	};
+
+	pending = tp_text_channel_get_pending_messages (TP_TEXT_CHANNEL (channel));
+	g_debug ("pending: %u", g_list_length (pending));
+	if (g_list_length (pending) > 0)
+	{
+		gchar *id;
+
+		id = channel_id (TP_CHANNEL (channel));
+
+		if (!messaging_menu_app_has_source (mmapp, id))
+		{
+			connection = tp_channel_borrow_connection (TP_CHANNEL (channel));
+			handle = tp_channel_get_initiator_handle (TP_CHANNEL (channel));
+			tp_connection_get_contacts_by_handle (connection, 1, &handle, 2, features, handle_contacts_pending_add_indicator_cb, channel, NULL, NULL);
+		}
+
+		g_free (id);
+	}
+        g_list_free (pending);
+	return FALSE;
+}
+
+static void
+pending_message_removed_cb (TpTextChannel *channel, TpMessage *message, gpointer user_data)
+{
+	gchar *id;
+	g_debug ("pending_message_removed_cb: %s", tp_channel_get_identifier (TP_CHANNEL (channel)));
+
+	id = channel_id (TP_CHANNEL (channel));
+	messaging_menu_app_remove_source (mmapp, id);
+	update_launcher (--n_sources);
+
+	g_free (id);
+}
+
+static void
+message_received_cb (TpTextChannel *channel, TpMessage *message, gpointer user_data)
+{
+	g_debug ("message_received_cb: %s", tp_channel_get_identifier (TP_CHANNEL (channel)));
+	g_timeout_add (100, check_pending_messages, channel);
+}
+
+static void
+handle_contacts_add_indicator_cb (TpConnection *connection, guint n_contacts, TpContact * const *contacts, guint n_failed, const TpHandle *failed, const GError *err, gpointer user_data, GObject *weak_object)
+{
+	gchar *id;
+	
+	TpChannelDispatchOperation *dispatch_op = TP_CHANNEL_DISPATCH_OPERATION (user_data);
+
+	TpChannel *channel;
+	GPtrArray *channels;
+	channels = tp_channel_dispatch_operation_borrow_channels (dispatch_op);
+	channel = g_ptr_array_index (channels, 0);
+
+	id = channel_id (channel);
+
+	if (messaging_menu_app_has_source (mmapp, id))
+	{
+		messaging_menu_app_set_source_time (mmapp, id, g_get_real_time ());
+	}
+	else
+	{
+		TpContact *contact;
+		const gchar *alias = "";
+		GFile *avatar_file;
+		GIcon *avatar;
+
+		contact = get_contact_from_channel (channel);
+		if (contact)
+		{
+			alias = tp_contact_get_alias (contact);
+			avatar_file = tp_contact_get_avatar_file (contact);
+			avatar = avatar_file ? g_file_icon_new (avatar_file) : NULL;
+		} else
+		{
+			alias = tp_channel_get_initiator_identifier (channel);
+		}
+
+		messaging_menu_app_append_source (mmapp, id, avatar, alias);
+		update_launcher (++n_sources);
+
+		g_clear_object (&avatar);
+	}
+
+	messaging_menu_app_draw_attention (mmapp, id);
+
+	g_hash_table_insert (dispatch_ops, g_strdup (id), dispatch_op);
+	g_signal_connect (dispatch_op, "invalidated", G_CALLBACK (dispatch_op_finished_cb), NULL);
+
+	g_free (id);
+}
+
+static void
+add_dispatch_operation_cb (TpSimpleApprover *self, TpAccount *account, TpConnection *connection,
+    GList *channels, TpChannelDispatchOperation *dispatch_op, TpAddDispatchOperationContext *context,
+    gpointer user_data)
+{
+	GList *itr;
+	TpChannel *channel;
+	TpHandle handle;
+
+	static TpContactFeature features[] = {
+		TP_CONTACT_FEATURE_ALIAS,
+		TP_CONTACT_FEATURE_AVATAR_DATA,
+	};
+
+	g_object_ref (dispatch_op);
+	for (itr = channels; itr != NULL; itr = g_list_next (itr))
+	{
+		channel = itr->data;
+		handle = tp_channel_get_initiator_handle (channel);
+		tp_connection_get_contacts_by_handle (connection, 1, &handle, 2, features, handle_contacts_add_indicator_cb, dispatch_op, NULL, NULL);
+	}
+	g_list_free(itr);
+	tp_add_dispatch_operation_context_accept (context);
+}
+
+static void
+observer_observe_channels_cb (TpSimpleObserver *self, TpAccount *account, TpConnection *connection,
+    GList *channels, TpChannelDispatchOperation *dispatch_op, GList *requests, TpObserveChannelsContext *context,
+    gpointer user_data)
+{
+	GList *itr;
+	TpChannel *channel;
+
+	for (itr = channels; itr != NULL; itr = g_list_next (itr))
+	{
+		channel = itr->data;
+		if (TP_IS_TEXT_CHANNEL (channel))
+		{
+			g_debug ("Handling text channel: %s", tp_channel_get_identifier (channel));
+
+			g_object_ref (channel);
+
+			g_signal_connect (channel, "message-received",
+				G_CALLBACK (message_received_cb), NULL);
+			g_signal_connect (channel, "pending-message-removed",
+				G_CALLBACK (pending_message_removed_cb), NULL);
+
+			check_pending_messages (TP_CHANNEL (channel));
+		}
+	}
+	g_list_free(itr);
+	tp_observe_channels_context_accept (context);
 }
 
 static void
@@ -776,13 +672,11 @@ static void
 contact_list_changed_cb (TpConnection *conn, GPtrArray *added, GPtrArray *removed, gpointer user_data)
 {
 	g_debug ("contact_list_changed_cb");
-	GTimeVal time;
+	gchar *id;
 	const gchar *alias = "";
 	guint i;
-        GFile *avatar = NULL;
-        GdkPixbuf *pixbuf = NULL;
-        GtkWidget *image = NULL;
-        GError *error = NULL;
+	GFile *avatar_file;
+	GIcon *avatar = NULL;
 
 	for (i = 0; i < added->len; i++)
 	{
@@ -792,46 +686,31 @@ contact_list_changed_cb (TpConnection *conn, GPtrArray *added, GPtrArray *remove
 			TpSubscriptionState state = tp_contact_get_publish_state (contact);
 			if (state == TP_SUBSCRIPTION_STATE_ASK)
 			{
-				alias = tp_contact_get_alias (contact);
-	
-				IndicateIndicator *indicator = NULL;
-				indicator = indicate_indicator_new ();
-				indicate_indicator_set_property (INDICATE_INDICATOR (indicator), "subtype", "im");
-				indicate_indicator_set_property (INDICATE_INDICATOR (indicator), "sender", alias);
-				indicate_indicator_set_property (INDICATE_INDICATOR (indicator), "draw-attention", "true");
-
-
 				g_debug ("Getting avatar for %s", alias);
 
-				avatar = tp_contact_get_avatar_file (contact);
+				id = contact_id (contact);
 
-				if (avatar)
+				if (messaging_menu_app_has_source (mmapp, id))
 				{
-					g_debug ("avatar");
-					pixbuf = gdk_pixbuf_new_from_file_at_scale (g_file_get_path (avatar), 22, 22, TRUE, &error);
+					messaging_menu_app_set_source_time (mmapp, id, g_get_real_time ());
 				}
-
-				if (error)
+				else
 				{
-					g_warning ("Failed to create pixbuf: %s", error->message);
-					g_error_free (error);
-				}
+					alias = tp_contact_get_alias (contact);
+					avatar_file = tp_contact_get_avatar_file (contact);
+					avatar = avatar_file ? g_file_icon_new (avatar_file) : NULL;
 
-				if (pixbuf)
-				{
-					indicate_gtk_indicator_set_property_icon (indicator, "icon", pixbuf);
-					g_object_unref(G_OBJECT(pixbuf));
-				}
+					messaging_menu_app_append_source (mmapp, id, avatar, alias);
+					messaging_menu_app_draw_attention (mmapp, id);
 
-				g_signal_connect (G_OBJECT (indicator), INDICATE_INDICATOR_SIGNAL_DISPLAY, G_CALLBACK (subscription_indicator_display), contact);
-	
-				g_get_current_time (&time);
-				indicate_indicator_set_property_time (INDICATE_INDICATOR (indicator), "time", &time);
+					n_sources++;
+
+					g_clear_object (&avatar);
+				}
 
 				g_debug ("contact_list_changed_cb: adding %s", alias);
 
-				indicate_indicator_show (INDICATE_INDICATOR (indicator));
-				g_hash_table_insert (indicators, indicator, contact);
+				g_free (id);
 			}
 		}
 	}
@@ -846,22 +725,11 @@ contact_list_changed_cb (TpConnection *conn, GPtrArray *added, GPtrArray *remove
 			contact_remove_from_contact_list (contact);
 		}
 
-		GHashTableIter iter;
-		gpointer data, indicator;
-		g_hash_table_iter_init (&iter, indicators);
-		while (g_hash_table_iter_next (&iter, &indicator, &data)) 
-		{
-			if (!TP_IS_CONTACT (data))
-				continue;
-			if (data == contact)
-			{
-				g_debug ("contact_list_changed_cb: found indicator for contact");
-				indicate_indicator_hide (INDICATE_INDICATOR (indicator));
-				g_hash_table_remove (indicators, indicator);
-				break;
-			}
-		}
+		messaging_menu_app_remove_source (mmapp, tp_contact_get_identifier (contact));
+		n_sources--;
 	}
+
+	update_launcher (n_sources);
 }
 
 
@@ -1012,7 +880,7 @@ observer_setup (TpAccountManager *tp_am)
 	g_return_if_fail (TP_IS_ACCOUNT_MANAGER (tp_am));
 
 	observer = tp_simple_observer_new_with_am (tp_am, TRUE, "IndicatorObserver",
-				FALSE, observer_add_dispatch_operation_cb, NULL, NULL);
+				FALSE, observer_observe_channels_cb, NULL, NULL);
 
 	/* contact text chat */
 	tp_base_client_take_observer_filter (observer, tp_asv_new (
@@ -1034,18 +902,93 @@ observer_setup (TpAccountManager *tp_am)
 	return observer;
 }
 
+static void
+presence_changed (TpAccountManager *manager,
+		  guint             presence,
+		  gchar            *status,
+		  gchar            *message,
+		  gpointer          user_data)
+{
+	MessagingMenuStatus mm_status;
+
+	switch (presence)
+	{
+	case TP_CONNECTION_PRESENCE_TYPE_AVAILABLE:
+		mm_status = MESSAGING_MENU_STATUS_AVAILABLE;
+		break;
+
+	case TP_CONNECTION_PRESENCE_TYPE_AWAY:
+	case TP_CONNECTION_PRESENCE_TYPE_EXTENDED_AWAY:
+		mm_status = MESSAGING_MENU_STATUS_AWAY;
+		break;
+
+	case TP_CONNECTION_PRESENCE_TYPE_HIDDEN:
+		mm_status = MESSAGING_MENU_STATUS_INVISIBLE;
+		break;
+
+	case TP_CONNECTION_PRESENCE_TYPE_BUSY:
+		mm_status = MESSAGING_MENU_STATUS_BUSY;
+		break;
+
+	default:
+	case TP_CONNECTION_PRESENCE_TYPE_UNKNOWN:
+	case TP_CONNECTION_PRESENCE_TYPE_OFFLINE:
+	case TP_CONNECTION_PRESENCE_TYPE_UNSET:
+		mm_status = MESSAGING_MENU_STATUS_OFFLINE;
+		break;
+	}
+
+	messaging_menu_app_set_status (mmapp, mm_status);
+}
+
+static void
+messaging_menu_status_changed (MessagingMenuApp    *mmapp,
+			       MessagingMenuStatus mm_status,
+			       gpointer            user_data)
+{
+	TpAccountManager *tp_am = user_data;
+	TpConnectionPresenceType status;
+
+	switch (mm_status)
+	{
+	case MESSAGING_MENU_STATUS_AVAILABLE:
+		status = TP_CONNECTION_PRESENCE_TYPE_AVAILABLE;
+		break;
+
+	case MESSAGING_MENU_STATUS_AWAY:
+		status = TP_CONNECTION_PRESENCE_TYPE_AWAY;
+		break;
+
+	case MESSAGING_MENU_STATUS_BUSY:
+		status = TP_CONNECTION_PRESENCE_TYPE_BUSY;
+		break;
+
+	case MESSAGING_MENU_STATUS_INVISIBLE:
+		status = TP_CONNECTION_PRESENCE_TYPE_HIDDEN;
+		break;
+
+	case MESSAGING_MENU_STATUS_OFFLINE:
+		status = TP_CONNECTION_PRESENCE_TYPE_OFFLINE;
+		break;
+
+	default:
+		g_assert_not_reached ();
+	}
+
+	tp_account_manager_set_all_requested_presences (tp_am, status, "", "");
+}
 
 int
 main (int argc, char **argv)
 {
 	GMainLoop *loop = NULL;
+	TpConnectionPresenceType presence;
 
 	g_type_init ();
 
 	gtk_init (&argc, &argv);
 	TpBaseClient *approver;
 	TpBaseClient *observer;
-	IndicateServer *indicate_server;
 
 	loop = g_main_loop_new (NULL, FALSE);
 
@@ -1064,8 +1007,23 @@ main (int argc, char **argv)
 		return 1;
 	}
 
+	mmapp = messaging_menu_app_new ("empathy.desktop");
+	messaging_menu_app_register (mmapp);
 
-	indicate_server = indicate_server_setup ();
+	g_signal_connect (mmapp, "activate-source",
+			  G_CALLBACK (message_source_activated), tp_am);
+	g_signal_connect (mmapp, "status-changed",
+			  G_CALLBACK (messaging_menu_status_changed), tp_am);
+
+	presence = tp_account_manager_get_most_available_presence (tp_am, NULL, NULL);
+	presence_changed (tp_am, presence, "", "", NULL);
+	g_signal_connect(tp_am, "most-available-presence-changed",
+			 G_CALLBACK (presence_changed), NULL);
+
+	dispatch_ops = g_hash_table_new_full (g_str_hash, g_str_equal,
+					      g_free, g_object_unref);
+
+	launcher = unity_launcher_entry_get_for_desktop_id ("empathy.desktop");
 
 	contact_list_setup (tp_am);
 
